@@ -1,9 +1,9 @@
 import subprocess
 import zlib
+from contextlib import contextmanager
 from io import BytesIO
 from os import fspath
 from pathlib import Path
-from typing import Dict
 
 import pytest
 from PIL import Image, ImageCms
@@ -14,6 +14,7 @@ from pikepdf import (
     Array,
     Dictionary,
     Name,
+    Operator,
     Pdf,
     PdfError,
     PdfImage,
@@ -32,30 +33,43 @@ from pikepdf.models.image import (
 # pylint: disable=redefined-outer-name
 
 
-def first_image_in(filename):
-    pdf = Pdf.open(filename)
-    pdfimagexobj = next(iter(pdf.pages[0].images.values()))
-    return pdfimagexobj, pdf
+@pytest.fixture
+def first_image_in(resources, request):
+    pdf = None
+
+    def opener(filename):
+        nonlocal pdf
+        pdf = Pdf.open(resources / filename)
+        pdfimagexobj = next(iter(pdf.pages[0].images.values()))
+        return pdfimagexobj, pdf
+
+    def closer():
+        if pdf:
+            pdf.close()
+
+    request.addfinalizer(closer)
+
+    return opener
 
 
 @pytest.fixture
-def congress(resources):
-    return first_image_in(resources / 'congress.pdf')
+def congress(first_image_in):
+    return first_image_in('congress.pdf')
 
 
 @pytest.fixture
-def sandwich(resources):
-    return first_image_in(resources / 'sandwich.pdf')
+def sandwich(first_image_in):
+    return first_image_in('sandwich.pdf')
 
 
 @pytest.fixture
-def jbig2(resources):
-    return first_image_in(resources / 'jbig2.pdf')
+def jbig2(first_image_in):
+    return first_image_in('jbig2.pdf')
 
 
 @pytest.fixture
-def trivial(resources):
-    return first_image_in(resources / 'pal-1bit-trivial.pdf')
+def trivial(first_image_in):
+    return first_image_in('pal-1bit-trivial.pdf')
 
 
 @pytest.fixture
@@ -63,14 +77,15 @@ def inline(resources):
     pdf = Pdf.open(resources / 'image-mono-inline.pdf')
     for operands, _command in parse_content_stream(pdf.pages[0]):
         if operands and isinstance(operands[0], PdfInlineImage):
-            return operands[0], pdf
+            yield operands[0], pdf
+    pdf.close()
 
 
 def test_image_from_nonimage(resources):
     pdf = Pdf.open(resources / 'congress.pdf')
-    resources = pdf.pages[0].Contents
+    contents = pdf.pages[0].Contents
     with pytest.raises(TypeError):
-        PdfImage(resources)
+        PdfImage(contents)
 
 
 def test_image(congress):
@@ -82,7 +97,7 @@ def test_image(congress):
 
 
 def test_imagemask(congress):
-    assert PdfImage(congress[0]).image_mask == False
+    assert not PdfImage(congress[0]).image_mask
 
 
 def test_imagemask_colorspace(trivial):
@@ -161,7 +176,7 @@ def test_lowlevel_replace_jpeg(congress, outdir):
 def test_inline(inline):
     iimage, pdf = inline
     assert iimage.width == 8
-    assert iimage.image_mask == False
+    assert not iimage.image_mask
     assert iimage.mode == 'RGB'
     assert iimage.is_inline
     assert iimage.colorspace == '/DeviceRGB'
@@ -172,8 +187,9 @@ def test_inline(inline):
     assert b'/Width 8' not in unparsed, "abbreviations expanded in inline image"
 
     cs = pdf.make_stream(unparsed)
-    for operands, _command in parse_content_stream(cs):
+    for operands, command in parse_content_stream(cs):
         if operands and isinstance(operands[0], PdfInlineImage):
+            assert command == Operator('INLINE IMAGE')
             reparsed_iim = operands[0]
             assert reparsed_iim == iimage
 
@@ -190,6 +206,7 @@ def test_bits_per_component_missing(congress):
         (1, 1, b'\xff', '/DeviceGray', 1),
         (1, 1, b'\xf0', '/DeviceGray', 8),
         (1, 1, b'\xff\x00\xff', '/DeviceRGB', 8),
+        (1, 1, b'\xff\x80\x40\x20', '/DeviceCMYK', 8),
     ],
 )
 def test_image_roundtrip(outdir, w, h, pixeldata, cs, bpc):
@@ -236,6 +253,8 @@ def test_image_roundtrip(outdir, w, h, pixeldata, cs, bpc):
             assert pim.mode == 'RGB'
         elif cs == '/DeviceGray' and bpc == 8:
             assert pim.mode == 'L'
+        elif cs == '/DeviceCMYK':
+            assert pim.mode == 'CMYK'
         elif bpc == 1:
             assert pim.mode == '1'
         assert not pim.palette
@@ -259,8 +278,8 @@ def test_image_roundtrip(outdir, w, h, pixeldata, cs, bpc):
         ('cmyk-jpeg.pdf', 8, ['/DCTDecode'], '.jpg', 'CMYK', 'JPEG'),
     ],
 )
-def test_direct_extract(resources, filename, bpc, filters, ext, mode, format_):
-    xobj, _pdf = first_image_in(resources / filename)
+def test_direct_extract(first_image_in, filename, bpc, filters, ext, mode, format_):
+    xobj, _pdf = first_image_in(filename)
     pim = PdfImage(xobj)
 
     assert pim.bits_per_component == bpc
@@ -309,9 +328,8 @@ def test_bool_in_inline_image():
 @pytest.mark.skipif(
     not PIL_features.check_codec('jpg_2000'), reason='no JPEG2000 codec'
 )
-def test_jp2(resources):
-    pdf = Pdf.open(resources / 'pike-jp2.pdf')
-    xobj = next(iter(pdf.pages[0].images.values()))
+def test_jp2(first_image_in):
+    xobj, pdf = first_image_in('pike-jp2.pdf')
     pim = PdfImage(xobj)
     assert isinstance(pim, PdfJpxImage)
 
@@ -368,8 +386,8 @@ def test_extract_direct_fails_nondefault_colortransform(congress):
         pim.extract_to(stream=bio)
 
 
-def test_icc_use(resources):
-    xobj, _pdf = first_image_in(resources / '1biticc.pdf')
+def test_icc_use(first_image_in):
+    xobj, _pdf = first_image_in('1biticc.pdf')
 
     pim = PdfImage(xobj)
     assert pim.mode == '1'
@@ -379,15 +397,15 @@ def test_icc_use(resources):
     assert pim.icc.profile.xcolor_space == 'GRAY'
 
 
-def test_icc_extract(resources):
-    xobj, _pdf = first_image_in(resources / 'aquamarine-cie.pdf')
+def test_icc_extract(first_image_in):
+    xobj, _pdf = first_image_in('aquamarine-cie.pdf')
 
     pim = PdfImage(xobj)
     assert pim.as_pil_image().info['icc_profile'] == pim.icc.tobytes()
 
 
-def test_icc_palette(resources):
-    xobj, _pdf = first_image_in(resources / 'pink-palette-icc.pdf')
+def test_icc_palette(first_image_in):
+    xobj, _pdf = first_image_in('pink-palette-icc.pdf')
     pim = PdfImage(xobj)
     assert pim.icc.profile.xcolor_space == 'RGB '  # with trailing space
     b = BytesIO()
@@ -404,8 +422,8 @@ def test_icc_palette(resources):
     assert pil_prf.tobytes() == pim.icc.tobytes()
 
 
-def test_stacked_compression(resources):
-    xobj, _pdf = first_image_in(resources / 'pike-flate-jp2.pdf')
+def test_stacked_compression(first_image_in):
+    xobj, _pdf = first_image_in('pike-flate-jp2.pdf')
 
     pim = PdfImage(xobj)
     assert pim.mode == 'RGB'
@@ -434,8 +452,8 @@ def test_ccitt_encodedbytealign(sandwich):
         pim.as_pil_image()
 
 
-def test_imagemagick_uses_rle_compression(resources):
-    xobj, _rle = first_image_in(resources / 'rle.pdf')
+def test_imagemagick_uses_rle_compression(first_image_in):
+    xobj, _rle = first_image_in('rle.pdf')
 
     pim = PdfImage(xobj)
     im = pim.as_pil_image()
@@ -472,8 +490,8 @@ def test_jbig2(jbig2):
 
 
 @needs_jbig2dec
-def test_jbig2_global(resources):
-    xobj, _pdf = first_image_in(resources / 'jbig2global.pdf')
+def test_jbig2_global(first_image_in):
+    xobj, _pdf = first_image_in('jbig2global.pdf')
     pim = PdfImage(xobj)
     im = pim.as_pil_image()
     assert im.size == (4000, 2864)
@@ -481,8 +499,8 @@ def test_jbig2_global(resources):
 
 
 @needs_jbig2dec
-def test_jbig2_global_palette(resources):
-    xobj, _pdf = first_image_in(resources / 'jbig2global.pdf')
+def test_jbig2_global_palette(first_image_in):
+    xobj, _pdf = first_image_in('jbig2global.pdf')
     xobj.ColorSpace = pikepdf.Array(
         [Name.Indexed, Name.DeviceRGB, 1, b'\x00\x00\x00\xff\xff\xff']
     )
@@ -492,8 +510,8 @@ def test_jbig2_global_palette(resources):
     assert im.getpixel((0, 0)) == 255  # Ensure loaded
 
 
-def test_jbig2_error(resources, monkeypatch):
-    xobj, _pdf = first_image_in(resources / 'jbig2global.pdf')
+def test_jbig2_error(first_image_in, monkeypatch):
+    xobj, _pdf = first_image_in('jbig2global.pdf')
     pim = PdfImage(xobj)
     monkeypatch.setattr(pikepdf.jbig2, 'jbig2dec_available', lambda: True)
 
@@ -507,8 +525,8 @@ def test_jbig2_error(resources, monkeypatch):
         pim.as_pil_image()
 
 
-def test_jbig2_too_old(resources, monkeypatch):
-    xobj, _pdf = first_image_in(resources / 'jbig2global.pdf')
+def test_jbig2_too_old(first_image_in, monkeypatch):
+    xobj, _pdf = first_image_in('jbig2global.pdf')
     pim = PdfImage(xobj)
 
     def run_version_override(subprocargs, *args, **kwargs):
@@ -523,8 +541,8 @@ def test_jbig2_too_old(resources, monkeypatch):
         pim.as_pil_image()
 
 
-def test_ccitt_icc(resources):
-    xobj, pdf = first_image_in(resources / 'sandwich.pdf')
+def test_ccitt_icc(first_image_in, resources):
+    xobj, pdf = first_image_in('sandwich.pdf')
 
     pim = PdfImage(xobj)
     assert pim.icc is None
@@ -552,8 +570,8 @@ def test_ccitt_icc(resources):
     assert Image.open(bio)
 
 
-def test_invalid_icc(resources):
-    xobj, _pdf = first_image_in(resources / 'pink-palette-icc.pdf')
+def test_invalid_icc(first_image_in):
+    xobj, _pdf = first_image_in('pink-palette-icc.pdf')
 
     cs = xobj.ColorSpace[1][1]  # [/Indexed [/ICCBased <stream>]]
     cs.write(b'foobar')  # corrupt the ICC profile
@@ -561,7 +579,7 @@ def test_invalid_icc(resources):
         UnsupportedImageTypeError, match="ICC profile corrupt or not readable"
     ):
         pim = PdfImage(xobj)
-        _icc = pim.icc
+        assert pim.icc is not None
 
 
 def test_dict_or_array_dict():
@@ -590,30 +608,206 @@ def test_dict_or_array_dict():
     assert pim.decode_parms[0].K == -1  # Check that array of dict is unpacked properly
 
 
+CMYK_RED = b'\x00\xc0\xc0\x15'
+CMYK_GREEN = b'\x90\x00\xc0\x15'
+CMYK_BLUE = b'\xc0\xa0\x00\x15'
+CMYK_PINK = b'\x04\xc0\x00\x15'
+
+CMYK_PALETTE = CMYK_RED + CMYK_GREEN + CMYK_BLUE + CMYK_PINK
+
+
 @pytest.mark.parametrize(
-    'base, hival, palette, expect_type',
+    'base, hival, palette, expect_type, expect_mode',
     [
-        (Name.DeviceGray, 4, b'\x00\x40\x80\xff', 'L'),
-        (Name.DeviceCMYK, 4, b'aaaabbbbccccdddd', 'CMYK'),
+        (Name.DeviceGray, 4, b'\x00\x40\x80\xff', 'L', 'P'),
+        (Name.DeviceCMYK, 4, CMYK_PALETTE, 'CMYK', 'P'),
     ],
 )
-def test_palette_nonrgb(base, hival, palette, expect_type):
+def test_palette_nonrgb(base, hival, palette, expect_type, expect_mode):
     pdf = pikepdf.new()
     imobj = Stream(
         pdf,
-        b'\x00\x01\x02\x03' * 4,
+        b'\x00\x01\x02\x03' * 16,
         BitsPerComponent=8,
         ColorSpace=Array([Name.Indexed, base, hival, palette]),
         Width=16,
-        Height=1,
+        Height=4,
         Type=Name.XObject,
         Subtype=Name.Image,
     )
     pim = pikepdf.PdfImage(imobj)
     assert pim.palette == (expect_type, palette)
+    pim.extract_to(stream=BytesIO())
+    # To view images:
+    # pim.extract_to(fileprefix=f'palette_nonrgb_{expect_type}')
+    assert pim.mode == expect_mode
 
 
 def test_extract_to_mutex_params(sandwich):
     pdfimage = PdfImage(sandwich[0])
     with pytest.raises(ValueError, match="Cannot set both"):
         pdfimage.extract_to(stream=BytesIO(), fileprefix='anything')
+
+
+def test_separation():
+    # Manually construct a 2"x1" document with a Separation
+    # colorspace that devices a single "spot" color channel named
+    # "LogoGreen". Define a conversion to standard CMYK that assigns
+    # CMYK equivalents. Copied example from PDF RM.
+    # LogoGreen is a teal-ish green. First panel is white to full green,
+    # second is green to full white. RGB ~= (31, 202, 113)
+    pdf = pikepdf.new()
+    pdf.add_blank_page(page_size=(144, 72))
+
+    # pikepdf does not interpret this - it is for the PDF viewer
+    # Explanation:
+    #   X is implicitly loaded to stack
+    #   dup: X X
+    #   0.84 mul: X 0.84X
+    #   exch: 0.84X X
+    #   0.00: 0.84X X 0.00
+    #   exch: 0.84X 0.00 X
+    #   dup: 0.84X 0.00 X X
+    #   0.44 mul: 0.84X 0.00 X 0.44X
+    #   exch: 0.84X 0.00 0.44X X
+    #   0.21mul: 0.84X 0.00 0.44X 0.21X
+    # X -> {0.84X, 0, 0.44X, 0.21X}
+    tint_transform_logogreen_to_cmyk = b'''
+    {
+        dup 0.84 mul
+        exch 0.00 exch dup 0.44 mul
+        exch 0.21 mul
+    }
+    '''
+
+    cs = Array(
+        [
+            Name.Separation,
+            Name.LogoGreen,
+            Name.DeviceCMYK,
+            Stream(
+                pdf,
+                tint_transform_logogreen_to_cmyk,
+                FunctionType=4,
+                Domain=[0.0, 1.0],
+                Range=[0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0],
+            ),
+        ]
+    )
+
+    def check_pim(imobj, idx):
+        pim = pikepdf.PdfImage(imobj)
+        assert pim.mode == 'Separation'
+        assert pim.is_separation
+        assert not pim.is_device_n
+        assert pim.indexed == idx
+        assert repr(pim)
+        with pytest.raises(pikepdf.models.image.HifiPrintImageNotTranscodableError):
+            pim.extract_to(stream=BytesIO())
+
+    imobj0 = Stream(
+        pdf,
+        bytes(range(0, 256)),
+        BitsPerComponent=8,
+        ColorSpace=cs,
+        Width=16,
+        Height=16,
+        Type=Name.XObject,
+        Subtype=Name.Image,
+    )
+    check_pim(imobj0, idx=False)
+
+    imobj1 = Stream(
+        pdf,
+        bytes(range(0, 256)),
+        BitsPerComponent=8,
+        ColorSpace=Array([Name.Indexed, cs, 255, bytes(range(255, -1, -1))]),
+        Width=16,
+        Height=16,
+        Type=Name.XObject,
+        Subtype=Name.Image,
+    )
+    check_pim(imobj1, idx=True)
+
+    pdf.pages[0].Contents = Stream(
+        pdf, b'72 0 0 72 0 0 cm /Im0 Do 1 0 0 1 1 0 cm /Im1 Do'
+    )
+    pdf.pages[0].Resources = Dictionary(XObject=Dictionary(Im0=imobj0, Im1=imobj1))
+    # pdf.save("separation.pdf")
+
+
+def test_devicen():
+    # Manually construct a 2"x1" document with a DeviceN
+    # colorspace that devices a single "spot" color channel named
+    # "Black". Define a conversion to standard CMYK that assigns
+    # C=0 M=0 Y=0 and lets black through. The result should appear as a
+    # gradient from white (top left) to black (bottom right) in the
+    # left cell, and black to white in the right cell.
+    pdf = pikepdf.new()
+    pdf.add_blank_page(page_size=(144, 72))
+
+    # Postscript function to map X -> CMYK={0, 0, 0, X}
+    # Explanation:
+    #   X is implicitly on the stack
+    #   0 0 0 <- load three zeros on to stack
+    #   stack contains: X 0 0 0
+    #   4 -1 roll <- roll stack 4 elements -1 times, meaning the order is reversed
+    #   stack contains: 0 0 0 X
+    # pikepdf currently does not interpret tint transformation functions. This
+    # is done so that the output test file can be checked in a PDF viewer.
+    tint_transform_k_to_cmyk = b'{0 0 0 4 -1 roll}'
+
+    cs = Array(
+        [
+            Name.DeviceN,
+            Array([Name.Black]),
+            Name.DeviceCMYK,
+            Stream(
+                pdf,
+                tint_transform_k_to_cmyk,
+                FunctionType=4,
+                Domain=[0.0, 1.0],
+                Range=[0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0],
+            ),
+        ]
+    )
+
+    def check_pim(imobj, idx):
+        pim = pikepdf.PdfImage(imobj)
+        assert pim.mode == 'DeviceN'
+        assert pim.is_device_n
+        assert not pim.is_separation
+        assert pim.indexed == idx
+        assert repr(pim)
+        with pytest.raises(pikepdf.models.image.HifiPrintImageNotTranscodableError):
+            pim.extract_to(stream=BytesIO())
+
+    imobj0 = Stream(
+        pdf,
+        bytes(range(0, 256)),
+        BitsPerComponent=8,
+        ColorSpace=cs,
+        Width=16,
+        Height=16,
+        Type=Name.XObject,
+        Subtype=Name.Image,
+    )
+    check_pim(imobj0, idx=False)
+
+    imobj1 = Stream(
+        pdf,
+        bytes(range(0, 256)),
+        BitsPerComponent=8,
+        ColorSpace=Array([Name.Indexed, cs, 255, bytes(range(255, -1, -1))]),
+        Width=16,
+        Height=16,
+        Type=Name.XObject,
+        Subtype=Name.Image,
+    )
+    check_pim(imobj1, idx=True)
+
+    pdf.pages[0].Contents = Stream(
+        pdf, b'72 0 0 72 0 0 cm /Im0 Do 1 0 0 1 1 0 cm /Im1 Do'
+    )
+    pdf.pages[0].Resources = Dictionary(XObject=Dictionary(Im0=imobj0, Im1=imobj1))
+    # pdf.save('devicen.pdf')
