@@ -112,7 +112,7 @@ def file_descriptor_is_open_for():
 
 def test_open_named_file_closed(resources, file_descriptor_is_open_for):
     path = resources / 'pal.pdf'
-    pdf = Pdf.open(path)
+    pdf = Pdf.open(path)  # no with clause
     assert file_descriptor_is_open_for(path)
 
     pdf.close()
@@ -124,7 +124,7 @@ def test_open_named_file_closed(resources, file_descriptor_is_open_for):
 def test_streamed_file_not_closed(resources, file_descriptor_is_open_for):
     path = resources / 'pal.pdf'
     stream = path.open('rb')
-    pdf = Pdf.open(stream)
+    pdf = Pdf.open(stream)  # no with clause
     assert file_descriptor_is_open_for(path)
 
     pdf.close()
@@ -181,17 +181,17 @@ def test_file_without_fileno(resources):
 
     # Confirm we automatically fallback to stream
     with Pdf.open(f, access_mode=pikepdf._qpdf.AccessMode.mmap) as pdf:
-        pass
+        assert pdf.filename
 
 
 def test_file_deny_mmap(resources, monkeypatch):
     import mmap
 
-    def raises_ioerror(*args, **kwargs):
-        raise IOError("This file is temporarily not mmap-able")
+    def raises_oserror(*args, **kwargs):
+        raise OSError("This file is temporarily not mmap-able")
 
-    monkeypatch.setattr(mmap, 'mmap', raises_ioerror)
-    with pytest.raises(IOError):
+    monkeypatch.setattr(mmap, 'mmap', raises_oserror)
+    with pytest.raises(OSError):
         Pdf.open(resources / 'pal.pdf', access_mode=pikepdf._qpdf.AccessMode.mmap_only)
 
     with Pdf.open(
@@ -248,5 +248,13 @@ def test_stop_iteration_on_close(resources):
 
     # Inspired by https://github.com/pikepdf/pikepdf/issues/114
     stream = StopIterationOnClose((resources / 'pal-1bit-trivial.pdf').read_bytes())
-    pdf = Pdf.open(stream)
+    pdf = Pdf.open(stream)  # no with clause
     pdf.close()
+
+
+def test_read_after_close(resources):
+    pdf = Pdf.open(resources / 'pal.pdf')  # no with clause
+    contents = pdf.pages[0].Contents
+    pdf.close()
+    with pytest.raises(PdfError, match="closed input source"):
+        contents.read_raw_bytes()
